const gulp = require('gulp');
const fse = require('fs-extra');
const path = require('path');
const ts = require('gulp-typescript');
const merge2 = require('merge2');
const through2 = require('through2');
const cwd = process.cwd();
const ES = 'es';
const LIB = 'lib';
const SRC = 'src';
const esDir = path.join(cwd, ES);
const libDir = path.join(cwd, LIB);
const tsConfigPath = path.join(cwd, 'tsconfig.json');
const tsCommonConfigPath = path.join(cwd, 'tsconfig.commonjs.json');
const tsESConfigPath = path.join(cwd, 'tsconfig.es.json');
const tsDefaultReporter = ts.reporter.defaultReporter();
const styles = [
	'src/**/*.less',
	'src/**/*.sass',
	'src/**/*.scss',
	'src/**/*.stylus',
	'src/**/*.css',
];
const sources = ['src/**/*.tsx', 'src/**/*.ts', 'src/**/*.jsx', 'src/**/*.js'];

const tsConfig = fse.readJsonSync(tsConfigPath) || {};
const tsEsConfig = fse.pathExistsSync(tsESConfigPath) ? fse.readJsonSync(tsESConfigPath) : {};
const tsCommonConfig = fse.pathExistsSync(tsCommonConfigPath)
	? fse.readJsonSync(tsCommonConfigPath)
	: {};

function isDynamicStylesheetFile(filePath) {
	return ['.less', '.stylus', '.sass', '.scss', '.css'].includes(path.extname(filePath));
}

function isTypescriptFile(filePath) {
	return ['.tsx', '.ts'].includes(path.extname(filePath));
}

function clean(dir) {
	fse.removeSync(dir);
}

gulp.task('clean', (done) => {
	clean(esDir);
	clean(libDir);
	done();
});

gulp.task('compile-with-es', (done) => {
	compile({
		esm: true,
		single: false,
		styles,
		sources,
	}).on('finish', done);
});

gulp.task('compile-with-lib', (done) => {
	compile({
		esm: false,
		single: false,
		styles,
		sources,
	}).on('finish', done);
});

function replace(src, dest) {
	return through2.obj(function (file, encoding, next) {
		file.path = file.path.replace(src, dest);
		this.push(file);
		next();
	});
}

function compileOtherFile({ esm, sources, single }) {
	const dir = esm ? esDir : libDir;
	const module = esm ? ES : LIB;
	const dest = single ? cwd : dir;
	const streams = [];
	if (sources.length > 0) {
		const source = gulp.src(sources).pipe(replace(SRC, module)).pipe(gulp.dest(dest));
		streams.push(source);
	}
	return merge2(streams.filter((s) => s));
}

function compileDynamicStyleSheet(options) {
	return compileOtherFile(options);
}

function compileTypescriptFile({ esm, sources, single }) {
	const dir = esm ? esDir : libDir;
	const module = esm ? ES : LIB;
	const dest = single ? cwd : dir;
	const streams = [];
	if (sources.length > 0) {
		const compileOptions = Object.assign(
			{
				noUnusedParameters: false,
				noUnusedLocals: false,
				strictNullChecks: true,
				jsx: 'preserve',
				moduleResolution: 'node',
				declaration: true,
				allowSyntheticDefaultImports: true,
			},
			tsConfig.compilerOptions,
			esm
				? {
						target: 'es6',
						module: 'esnext',
				  }
				: {
						target: 'es5',
						module: 'commonjs',
				  },
			esm ? tsEsConfig.compilerOptions : tsCommonConfig.compilerOptions
		);

		let sourceStream = gulp.src(sources);
		let hasError = false;
		const tsResult = sourceStream
			.pipe(
				ts(compileOptions, {
					error(e) {
						hasError = true;
						tsDefaultReporter.error(e);
					},
					finish: tsDefaultReporter.finish,
				})
			)
			.on('error', () => {});
		const tsFilesStream = tsResult.js.pipe(replace(SRC, module)).pipe(gulp.dest(dest));
		const tsd = tsResult.dts.pipe(replace(SRC, module)).pipe(gulp.dest(dest));

		function check() {
			if (hasError) {
				// process.exit(1);
			}
		}

		tsResult.on('finish', check);
		tsResult.on('end', check);
		streams.push(tsFilesStream);
		streams.push(tsd);
	}
	return merge2(streams.filter((s) => s));
}

function compile({ esm, styles, sources, single }) {
	const dir = esm ? esDir : libDir;
	const module = esm ? ES : LIB;
	const dest = single ? cwd : dir;
	const streams = [];
	if (styles.length > 0) {
		streams.push(compileDynamicStyleSheet({ esm, single, sources: styles }));
	}
	if (sources.length > 0) {
		streams.push(compileTypescriptFile({ esm, single, sources }));
	}
	return merge2(streams.filter((s) => s));
}

gulp.task(
	'compile',
	gulp.series('clean', gulp.parallel(['compile-with-es', 'compile-with-lib']), (done) => {
		done();
	})
);

function deleteFile(filePath) {
	const extname = path.extname(filePath);
	const filename = path.basename(filePath, extname);
	const dirname = path.dirname(filePath).replace('src', '');

	function deleteTypescriptFile(fielPath) {
		const paths = [
			path.join(libDir, dirname, `${filename}.js`),
			path.join(libDir, dirname, `${filename}.d.ts`),

			path.join(esDir, dirname, `${filename}.js`),
			path.join(esDir, dirname, `${filename}.d.ts`),
		];
		paths.forEach((path) => {
			fse.remove(path);
		});
	}

	function deleteDynamicStylesheetFile(filePath) {
		deleteOtherFile(filePath);
	}

	function deleteOtherFile(filePath) {
		const paths = [filePath.replace('src', LIB), filePath.replace('src', ES)];

		paths.forEach((path) => {
			fse.remove(path);
		});
	}

	if (isTypescriptFile(filePath)) {
		deleteTypescriptFile(filePath);
		return;
	}

	if (isDynamicStylesheetFile(filePath)) {
		deleteDynamicStylesheetFile(filePath);
		return;
	}

	deleteOtherFile(filePath);
}

function watch(done) {
	const watcher = gulp.watch(['src/**']);

	watcher.on('unlink', function (path, stats) {
		console.log(`File ${path} was removed`);
		deleteFile(path);
	});

	watcher.on('add', function (path, stats) {
		console.log(`File ${path} was added`);
		compileSingleFile(path);
	});

	watcher.on('change', function (path, stats) {
		console.log(`File ${path} was change`);
		compileSingleFile(path);
	});
}

function compileSingleFile(path) {
	const esOptions = { esm: true, single: true, sources: [path] };
	const libOptions = { esm: false, single: true, sources: [path] };

	if (isTypescriptFile(path)) {
		compileTypescriptFile(esOptions);
		compileTypescriptFile(libOptions);
		return;
	}

	if (isDynamicStylesheetFile(path)) {
		compileDynamicStyleSheet(esOptions);
		compileDynamicStyleSheet(libOptions);
		return;
	}

	compileOtherFile(esOptions);
	compileOtherFile(libOptions);
	return;
}

gulp.task(
	'dev',
	gulp.series('clean', gulp.parallel(['compile-with-es', 'compile-with-lib']), watch)
);

gulp.task('build', gulp.series('clean', gulp.parallel(['compile-with-es', 'compile-with-lib'])));

// 使用 --require 时是前置加载，因此某种场景下可能需要后置加载
// require(path.resolve(process.cwd(), 'gulpfile'));
